class EventError(Exception):
    """A particular exception to throw from
    :py:class:`~symphony.bdk.core.service.datafeed.real_time_event_listener.RealTimeEventListener` implementations
    to explicitly indicate that the event processing has failed and that all the events received in
    the datafeed read events call should be re-queued.
    They will eventually get redispatched after some time (30s by default). Only supported by DFv2.
    """



"""Module handling authentication part of the configuration.
"""
from symphony.bdk.core.config.model.bdk_rsa_key_config import BdkRsaKeyConfig
from symphony.bdk.core.config.model.bdk_certificate_config import BdkCertificateConfig


class BdkAuthenticationConfig:
    """Class that contains the bot authentication configuration:
    - either private key configuration for RSA authentication
    - or certificate configuration for certificate authentication.
    """

    def __init__(self, private_key_config=None, certificate_config=None):
        self.private_key = BdkRsaKeyConfig(
            **private_key_config) if private_key_config is not None else BdkRsaKeyConfig()
        self.certificate = BdkCertificateConfig(
            **certificate_config) if certificate_config is not None else BdkCertificateConfig()

    def is_authentication_configured(self) -> bool:
        """Check if either RSA or certificate authentication is configured

        :return: True if RSA or certificate authentication is configured
        """
        return self.is_rsa_configuration_valid() or self.is_certificate_configuration_valid()

    def is_rsa_authentication_configured(self) -> bool:
        """Check if the RSA authentication is configured

        :return True if the RSA authentication is configured
        """
        return self.private_key.is_configured()

    def is_rsa_configuration_valid(self) -> bool:
        """Check of the RSA authentication is valid

        If both of private_key path and content, the configuration is invalid.
        :return: True if the the RSA key valid
        """
        return self.private_key.is_configured() and self.private_key.is_valid()

    def is_certificate_configuration_valid(self) -> bool:
        """Check of the certificate authentication is valid

        If both of certificate key path and content, the configuration is invalid.
        :return: True if the the certificate key valid
        """
        return self.certificate.is_valid()



"""Module containing all configuration related exception.
"""


class BdkConfigError(Exception):
    """Exception class raised when configuration is invalid.
    """


class BotNotConfiguredError(Exception):
    """Thrown when the bot configuration is not specified."""

    def __init__(self, message="Bot (service account) credentials have not been configured."):
        super().__init__(message)



"""
    Authenticator API

    For bots and other on-premise processes to authenticate. Once authenticated, the bot will be able to use the methods described in serviceAPI.yaml and agentAPI.yaml.  Connections to the servers will be over client authenticated TLS, the servers for this API will perform the authentication by inspecting the certificate presented by the SSLSocketClient.  There will be two implementations of this API, one on your Pod and one on the Key Manager. In order to fully authenticate, an API client will have to call both of these implementations and pass both of the session tokens returned as headers in all subsequent requests to the Symphony API.   # noqa: E501

    The version of the OpenAPI document: 20.12.1
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from symphony.bdk.gen.api_client import ApiClient, Endpoint as _Endpoint
from symphony.bdk.gen.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from symphony.bdk.gen.auth_model.error import Error
from symphony.bdk.gen.auth_model.pod_certificate import PodCertificate


class CertificatePodApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __v1_app_pod_certificate_get(
            self,
            **kwargs
        ):
            """Retrieve the certificate that can be use to validate the JWT token obtain through the extension application authentication flow.   # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = auth_api.v1_app_pod_certificate_get(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                PodCertificate
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        self.v1_app_pod_certificate_get = _Endpoint(
            settings={
                'response_type': (PodCertificate,),
                'auth': [],
                'endpoint_path': '/v1/app/pod/certificate',
                'operation_id': 'v1_app_pod_certificate_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_app_pod_certificate_get
        )



TRUST_STORE = "trustStore"


class BdkSslConfig:
    """Class containing the SSL configuration.
    self.trust_store_path should be the path to a file of concatenated CA certificates in PEM format."""

    def __init__(self, config):
        """

        :param config: the dict containing the SSL specific configuration.
        """
        self.trust_store_path = None
        if config is not None and TRUST_STORE in config:
            self.trust_store_path = config[TRUST_STORE].get("path")



from typing import Union, List, Tuple, IO

from symphony.bdk.core.auth.auth_session import AuthSession
from symphony.bdk.core.config.model.bdk_retry_config import BdkRetryConfig
from symphony.bdk.core.service.message.model import Message
from symphony.bdk.core.service.message.multi_attachments_messages_api import MultiAttachmentsMessagesApi
from symphony.bdk.gen.agent_api.attachments_api import AttachmentsApi
from symphony.bdk.gen.agent_model.v4_import_response import V4ImportResponse
from symphony.bdk.gen.agent_model.v4_imported_message import V4ImportedMessage
from symphony.bdk.gen.agent_model.v4_message import V4Message
from symphony.bdk.gen.agent_model.v4_message_blast_response import V4MessageBlastResponse
from symphony.bdk.gen.agent_model.v4_message_import_list import V4MessageImportList
from symphony.bdk.gen.pod_api.default_api import DefaultApi
from symphony.bdk.gen.pod_api.message_api import MessageApi
from symphony.bdk.gen.pod_api.message_suppression_api import MessageSuppressionApi
from symphony.bdk.gen.pod_api.pod_api import PodApi
from symphony.bdk.gen.pod_api.streams_api import StreamsApi
from symphony.bdk.gen.pod_model.message_metadata_response import MessageMetadataResponse
from symphony.bdk.gen.pod_model.message_receipt_detail_response import MessageReceiptDetailResponse
from symphony.bdk.gen.pod_model.message_status import MessageStatus
from symphony.bdk.gen.pod_model.message_suppression_response import MessageSuppressionResponse
from symphony.bdk.gen.pod_model.stream_attachment_item import StreamAttachmentItem

from symphony.bdk.core.retry import retry


class OboMessageService:
    """Class exposing OBO enabled endpoints for message management, e.g. send a message."""

    def __init__(self, messages_api: MultiAttachmentsMessagesApi, message_suppression_api: MessageSuppressionApi,
                 auth_session: AuthSession,
                 retry_config: BdkRetryConfig):
        self._messages_api = messages_api
        self._message_suppression_api = message_suppression_api
        self._auth_session = auth_session
        self._retry_config = retry_config

    async def send_message(
            self,
            stream_id: str,
            message: Union[str, Message],
            data=None,
            version: str = "",
            attachment: List[Union[IO, Tuple[IO, IO]]] = None
    ) -> V4Message:
        """Send a message to an existing stream.
        See: `Create Message <https://developers.symphony.com/restapi/reference#create-message-v4>`_

        :param stream_id: The ID of the stream to send the message to.
        :param message: a :py:class:`Message` instance or a string containing the MessageML content to be sent.
          If it is a :py:class:`Message` instance, other parameters will be ignored.
          If it is a string, ``<messageML>`` tags can be omitted.
        :param data: an object (e.g. dict) that will be serialized into JSON using ``json.dumps``.
        :param version: Optional message version in the format "major.minor".
          If empty, defaults to the latest supported version.
        :param attachment: One or more files (opened in binary or text mode) to be sent along with the message.
          Elements of the list can be attachment files or tuples of ``(attachment, preview)``.
          If one attachment has a preview, then all attachments must have a preview.
          The limit is set to 30Mb total size; also, it is recommended not to exceed 25 files.

        :return: a V4Message object containing the details of the sent message.
        """
        message_object = message if isinstance(message, Message) else Message(content=message, data=data,
                                                                              version=version, attachments=attachment)
        return await self._send_message(stream_id, message_object.content, message_object.data, message_object.version,
                                        message_object.attachments, message_object.previews)

    @retry
    async def _send_message(
            self,
            stream_id: str,
            message: str,
            data: str = "",
            version: str = "",
            attachment: List[IO] = None,
            preview: List[IO] = None
    ) -> V4Message:
        """Send a message to an existing stream.
        See: `Create Message <https://developers.symphony.com/restapi/reference#create-message-v4>`_

        :param stream_id: The ID of the stream to send the message to.
        :param message: The MessageML content to be sent.
        :param data: JSON data representing the objects contained in the message.
        :param version: Optional message version in the format "major.minor".
                        If empty, defaults to the latest supported version.
        :param attachment: One or more files to be sent along with the message.
                        The limit is set to 30Mb total size; also, it is recommended not to exceed 25 files.
        :param preview: Previews of the attachments which are sent along with the message.

        :return: a V4Message object containing the details of the sent message.

        """
        params = {
            "sid": stream_id,
            "session_token": await self._auth_session.session_token,
            "key_manager_token": await self._auth_session.key_manager_token,
            "message": message,
            "data": data,
            "version": version
        }
        if attachment is not None:
            params["attachment"] = attachment if isinstance(attachment, list) else [attachment]
        if preview is not None:
            params["preview"] = preview if isinstance(preview, list) else [preview]

        return await self._messages_api.v4_stream_sid_multi_attachment_message_create_post(**params)

    @retry
    async def suppress_message(
            self,
            message_id: str
    ) -> MessageSuppressionResponse:
        """Suppresses a message, preventing its contents from being displayed to users.
        See: `Suppress Message <https://developers.symphony.com/restapi/reference#suppress-message>`_

        :param message_id: Message ID of the message to be suppressed.

        :return: a MessageSuppressionResponse instance containing information about the message suppression.

        """
        params = {
            "id": message_id,
            "session_token": await self._auth_session.session_token
        }
        return await self._message_suppression_api.v1_admin_messagesuppression_id_suppress_post(**params)


class MessageService(OboMessageService):
    """Service class for managing messages.
    """

    def __init__(self, messages_api: MultiAttachmentsMessagesApi,
                 message_api: MessageApi,
                 message_suppression_api: MessageSuppressionApi,
                 streams_api: StreamsApi,
                 pod_api: PodApi,
                 attachment_api: AttachmentsApi,
                 default_api: DefaultApi,
                 auth_session: AuthSession,
                 retry_config: BdkRetryConfig):
        super().__init__(messages_api, message_suppression_api, auth_session, retry_config)
        self._message_api = message_api
        self._message_suppression_api = message_suppression_api
        self._streams_api = streams_api
        self._pod_api = pod_api
        self._attachment_api = attachment_api
        self._default_api = default_api

    @retry
    async def list_messages(
            self,
            stream_id: str,
            since: int = 0,
            skip: int = 0,
            limit: int = 50
    ) -> [V4Message]:
        """Get messages from an existing stream. Additionally returns any attachments associated with the message.
        See: `Messages <https://developers.symphony.com/restapi/reference#messages-v4>`_

        :param stream_id: The stream where to look for messages
        :param since: Timestamp of the earliest possible date of the first message returned.
        :param skip: Number of messages to skip. Default: 0
        :param limit: Maximum number of messages to return. Default: 50

        :return: the list of matching messages in the stream.

        """
        params = {
            "sid": stream_id,
            "since": since,
            "session_token": await self._auth_session.session_token,
            "key_manager_token": await self._auth_session.key_manager_token,
            "skip": skip,
            "limit": limit
        }
        message_list = await self._messages_api.v4_stream_sid_message_get(**params)
        return message_list.value

    async def blast_message(
            self,
            stream_ids: List[str],
            message: Union[str, Message],
            data=None,
            version: str = "",
            attachment: List[Union[IO, Tuple[IO, IO]]] = None
    ) -> V4MessageBlastResponse:
        """Send a message to multiple existing streams.
        See: `Blast Message <https://developers.symphony.com/restapi/reference#blast-message>`_

        :param stream_ids: The list of stream IDs to send the message to
        :param message: a :py:class:`Message` instance or a string containing the MessageML content to be sent.
          If it is a :py:class:`Message` instance, other parameters will be ignored.
          If it is a string, ``<messageML>`` tags can be omitted.
        :param data: an object (e.g. dict) that will be serialized into JSON using ``json.dumps``.
        :param version: Optional message version in the format "major.minor".
          If empty, defaults to the latest supported version.
        :param attachment: One or more files (opened in binary or text mode) to be sent along with the message.
          Elements of the list can be attachment files or tuples of ``(attachment, preview)``.
          If one attachment has a preview, then all attachments must have a preview.
          The limit is set to 30Mb total size; also, it is recommended not to exceed 25 files.
        :return:
        """
        message_object = message if isinstance(message, Message) else Message(content=message, data=data,
                                                                              version=version, attachments=attachment)
        return await self._blast_message(stream_ids, message_object.content, message_object.data,
                                         message_object.version, message_object.attachments, message_object.previews)

    @retry
    async def _blast_message(
            self,
            stream_ids: [str],
            message: str,
            data: str = "",
            version: str = "",
            attachment: List[IO] = None,
            preview: List[IO] = None
    ) -> V4MessageBlastResponse:
        """Send a message to multiple existing streams.
        See: `Blast Message <https://developers.symphony.com/restapi/reference#blast-message>`_

        :param stream_ids: The list of stream IDs to send the message to
        :param message: The messageML content to be sent.
        :param data: JSON data representing the objects contained in the message.
        :param version: Optional message version in the format "major.minor".
                        If empty, defaults to the latest supported version.
        :param attachment: One or more files to be sent along with the message.
                        The limit is set to 30Mb total size; also, it is recommended not to exceed 25 files.
        :param preview: Previews of the attachments which are sent along with the message.

        :return: a V4MessageBlastResponse object containing the details of the sent messages

        """
        params = {
            "sids": stream_ids,
            "session_token": await self._auth_session.session_token,
            "key_manager_token": await self._auth_session.key_manager_token,
            "message": message,
            "data": data,
            "version": version
        }
        if attachment is not None:
            params["attachment"] = attachment if isinstance(attachment, list) else [attachment]
        if preview is not None:
            params["preview"] = preview if isinstance(preview, list) else [preview]

        return await self._messages_api.v4_multi_attachment_message_blast_post(**params)

    @retry
    async def import_messages(
            self,
            messages: List[V4ImportedMessage]
    ) -> [V4ImportResponse]:
        """Imports a list of messages to Symphony.
        See: `Import Message <https://developers.symphony.com/restapi/reference#import-message-v4>`_

        :param messages: List of messages to import.

        :return: The list of imported messages

        """
        params = {
            "message_list": V4MessageImportList(value=messages),
            "session_token": await self._auth_session.session_token,
            "key_manager_token": await self._auth_session.key_manager_token
        }
        import_response_list = await self._messages_api.v4_message_import_post(**params)
        return import_response_list.value

    async def get_attachment(
            self,
            stream_id: str,
            message_id: str,
            attachment_id: str
    ) -> str:
        """Downloads the attachment body by the stream ID, message ID and attachment ID.
        See: `Attachment <https://developers.symphony.com/restapi/reference#attachment>`_

        :param stream_id: The stream ID where to look for the attachment.
        :param message_id: The ID of the message containing the attachment.
        :param attachment_id: The ID of the attachment

        :return: a byte array of attachment encoded in base 64.

        """
        params = {
            "sid": stream_id,
            "file_id": attachment_id,
            "message_id": message_id,
            "session_token": await self._auth_session.session_token,
            "key_manager_token": await self._auth_session.key_manager_token
        }
        return await self._attachment_api.v1_stream_sid_attachment_get(**params)

    async def get_message_status(
            self,
            message_id: str
    ) -> MessageStatus:
        """Get the status of a particular message, i.e the list of users who the message was sent to,
        delivered to and the list of users who read the message.
        See: `Message Status <https://developers.symphony.com/restapi/reference#message-status>`_

        :param message_id: MessageId the ID of the message to be checked

        :return: Status of the given message

        """
        params = {
            "mid": message_id,
            "session_token": await self._auth_session.session_token
        }
        return await self._message_api.v1_message_mid_status_get(**params)

    @retry
    async def get_attachment_types(self) -> List[str]:
        """Retrieves a list of supported file extensions for attachments.
        See: `Attachment Types <https://developers.symphony.com/restapi/reference#attachment-types>`_

        :return: a list of String containing all allowed file extensions for attachments.

        """
        params = {
            "session_token": await self._auth_session.session_token
        }
        type_list = await self._pod_api.v1_files_allowed_types_get(**params)
        return type_list.value

    @retry
    async def get_message(
            self,
            message_id: str
    ) -> V4Message:
        """Retrieves the details of a message given its message ID.
        See: `Get Message <https://developers.symphony.com/restapi/reference#get-message-v1>`_

        :param message_id: MessageId the ID of the message to be retrieved.

        :return: a V4Message containing the message's details.

        """
        params = {
            "id": message_id,
            "session_token": await self._auth_session.session_token,
            "key_manager_token": await self._auth_session.key_manager_token
        }
        return await self._messages_api.v1_message_id_get(**params)

    @retry
    async def list_attachments(
            self,
            stream_id: str,
            since: int = None,
            to: int = None,
            limit: int = 50,
            sort_dir: str = "ASC"
    ) -> List[StreamAttachmentItem]:
        """List attachments in a particular stream.
        See: `List Attachments <https://developers.symphony.com/restapi/reference#list-attachments>`_

        :param stream_id: The stream ID where to look for the attachments
        :param since: Timestamp of the first required attachment.
        :param to: Timestamp of the last required attachment.
        :param limit: Maximum number of attachments to return.
                    This optional value defaults to 50 and should be between 0 and 100.
        :param sort_dir: Attachment date sort direction : ASC or DESC. Default: ASC.

        :return: the list of attachments in the stream.

        """
        params = {
            "sid": stream_id,
            "session_token": await self._auth_session.session_token,
            "limit": limit,
            "sort_dir": sort_dir
        }
        if since is not None:
            params["since"] = since
        if to is not None:
            params["to"] = to
        attachment_list = await self._streams_api.v1_streams_sid_attachments_get(**params)
        return attachment_list.value

    @retry
    async def list_message_receipts(
            self,
            message_id: str
    ) -> MessageReceiptDetailResponse:
        """Fetches receipts details from a specific message.
        See: `List Message Receipts <https://developers.symphony.com/restapi/reference#list-message-receipts>`_

        :param message_id: MessageId the ID of the message to get receipt details from.

        :return:a MessageReceiptDetailResponse object holding all receipt information.

        """
        params = {
            "message_id": message_id,
            "session_token": await self._auth_session.session_token
        }
        return await self._default_api.v1_admin_messages_message_id_receipts_get(**params)

    @retry
    async def get_message_relationships(
            self,
            message_id: str
    ) -> MessageMetadataResponse:
        """Gets the message metadata relationship.
        This API allows users to track the relationship between a message and all the forwards and replies of that
        message.
        See: `Message Metadata <https://developers.symphony.com/restapi/reference#message-metadata-relationship>`_


        :param message_id: the ID of the message to get relationships from.
        :return: a MessageMetadataResponse object holding information about the current message relationships (parent,
            replies, forwards and form replies).
        """
        params = {
            "message_id": message_id,
            "session_token": await self._auth_session.session_token,
            "user_agent": ""
        }
        return await self._default_api.v1_admin_messages_message_id_metadata_relationships_get(**params)



"""
    Agent API

    This document refers to Symphony API calls to send and receive messages and content. They need the on-premise Agent installed to perform decryption/encryption of content.  - sessionToken and keyManagerToken can be obtained by calling the authenticationAPI on the symphony back end and the key manager respectively. Refer to the methods described in authenticatorAPI.yaml. - Actions are defined to be atomic, ie will succeed in their entirety or fail and have changed nothing. - If it returns a 40X status then it will have sent no message to any stream even if a request to aome subset of the requested streams would have succeeded. - If this contract cannot be met for any reason then this is an error and the response code will be 50X. - MessageML is a markup language for messages. See reference here: https://rest-api.symphony.com/docs/messagemlv2 - **Real Time Events**: The following events are returned when reading from a real time messages and events stream (\"datafeed\"). These events will be returned for datafeeds created with the v5 endpoints. To know more about the endpoints, refer to Create Messages/Events Stream and Read Messages/Events Stream. Unless otherwise specified, all events were added in 1.46.   # noqa: E501

    The version of the OpenAPI document: 20.12.0
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from symphony.bdk.gen.api_client import ApiClient, Endpoint as _Endpoint
from symphony.bdk.gen.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from symphony.bdk.gen.agent_model.agent_info import AgentInfo
from symphony.bdk.gen.agent_model.base_signal import BaseSignal
from symphony.bdk.gen.agent_model.channel_subscriber_response import ChannelSubscriberResponse
from symphony.bdk.gen.agent_model.channel_subscription_response import ChannelSubscriptionResponse
from symphony.bdk.gen.agent_model.signal import Signal
from symphony.bdk.gen.agent_model.signal_list import SignalList
from symphony.bdk.gen.agent_model.success_response import SuccessResponse
from symphony.bdk.gen.agent_model.v2_error import V2Error


class SignalsApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __v1_info_get(
            self,
            **kwargs
        ):
            """Get information about the Agent  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v1_info_get(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                AgentInfo
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        self.v1_info_get = _Endpoint(
            settings={
                'response_type': (AgentInfo,),
                'auth': [],
                'endpoint_path': '/v1/info',
                'operation_id': 'v1_info_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_info_get
        )

        def __v1_signals_create_post(
            self,
            session_token,
            signal,
            **kwargs
        ):
            """Create a signal.  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v1_signals_create_post(session_token, signal, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                signal (BaseSignal): Signal definition.

            Keyword Args:
                key_manager_token (str): Key Manager authentication token.. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Signal
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['signal'] = \
                signal
            return self.call_with_http_info(**kwargs)

        self.v1_signals_create_post = _Endpoint(
            settings={
                'response_type': (Signal,),
                'auth': [],
                'endpoint_path': '/v1/signals/create',
                'operation_id': 'v1_signals_create_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'signal',
                    'key_manager_token',
                ],
                'required': [
                    'session_token',
                    'signal',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'signal':
                        (BaseSignal,),
                    'key_manager_token':
                        (str,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'key_manager_token': 'keyManagerToken',
                },
                'location_map': {
                    'session_token': 'header',
                    'signal': 'body',
                    'key_manager_token': 'header',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_signals_create_post
        )

        def __v1_signals_id_delete_post(
            self,
            session_token,
            id,
            **kwargs
        ):
            """Delete a signal.  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v1_signals_id_delete_post(session_token, id, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                id (str): The id of the signal.

            Keyword Args:
                key_manager_token (str): Key Manager authentication token.. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['id'] = \
                id
            return self.call_with_http_info(**kwargs)

        self.v1_signals_id_delete_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/signals/{id}/delete',
                'operation_id': 'v1_signals_id_delete_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'id',
                    'key_manager_token',
                ],
                'required': [
                    'session_token',
                    'id',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'id':
                        (str,),
                    'key_manager_token':
                        (str,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'id': 'id',
                    'key_manager_token': 'keyManagerToken',
                },
                'location_map': {
                    'session_token': 'header',
                    'id': 'path',
                    'key_manager_token': 'header',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_signals_id_delete_post
        )

        def __v1_signals_id_get_get(
            self,
            session_token,
            id,
            **kwargs
        ):
            """Get details of the requested signal.  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v1_signals_id_get_get(session_token, id, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                id (str): The ID of the signal to display. 

            Keyword Args:
                key_manager_token (str): Key Manager authentication token.. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Signal
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['id'] = \
                id
            return self.call_with_http_info(**kwargs)

        self.v1_signals_id_get_get = _Endpoint(
            settings={
                'response_type': (Signal,),
                'auth': [],
                'endpoint_path': '/v1/signals/{id}/get',
                'operation_id': 'v1_signals_id_get_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'id',
                    'key_manager_token',
                ],
                'required': [
                    'session_token',
                    'id',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'id':
                        (str,),
                    'key_manager_token':
                        (str,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'id': 'id',
                    'key_manager_token': 'keyManagerToken',
                },
                'location_map': {
                    'session_token': 'header',
                    'id': 'path',
                    'key_manager_token': 'header',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_signals_id_get_get
        )

        def __v1_signals_id_subscribe_post(
            self,
            session_token,
            id,
            **kwargs
        ):
            """Subscribe to a Signal.  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v1_signals_id_subscribe_post(session_token, id, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                id (str): The id of the signal.

            Keyword Args:
                key_manager_token (str): Key Manager authentication token.. [optional]
                pushed (bool): Prevent the user to unsubscribe (only for bulk subscription). [optional]
                users ([int]): UserIds to subscribe (only for bulk subscription). [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ChannelSubscriptionResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['id'] = \
                id
            return self.call_with_http_info(**kwargs)

        self.v1_signals_id_subscribe_post = _Endpoint(
            settings={
                'response_type': (ChannelSubscriptionResponse,),
                'auth': [],
                'endpoint_path': '/v1/signals/{id}/subscribe',
                'operation_id': 'v1_signals_id_subscribe_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'id',
                    'key_manager_token',
                    'pushed',
                    'users',
                ],
                'required': [
                    'session_token',
                    'id',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'id':
                        (str,),
                    'key_manager_token':
                        (str,),
                    'pushed':
                        (bool,),
                    'users':
                        ([int],),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'id': 'id',
                    'key_manager_token': 'keyManagerToken',
                    'pushed': 'pushed',
                },
                'location_map': {
                    'session_token': 'header',
                    'id': 'path',
                    'key_manager_token': 'header',
                    'pushed': 'query',
                    'users': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_signals_id_subscribe_post
        )

        def __v1_signals_id_subscribers_get(
            self,
            session_token,
            id,
            **kwargs
        ):
            """Get the subscribers of a signal  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v1_signals_id_subscribers_get(session_token, id, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                id (str): The id of the signal.

            Keyword Args:
                key_manager_token (str): Key Manager authentication token.. [optional]
                skip (int): No. of results to skip.. [optional] if omitted the server will use the default value of 0
                limit (int): Max No. of subscribers to return. If no value is provided, 100 is the default.. [optional] if omitted the server will use the default value of 100
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ChannelSubscriberResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['id'] = \
                id
            return self.call_with_http_info(**kwargs)

        self.v1_signals_id_subscribers_get = _Endpoint(
            settings={
                'response_type': (ChannelSubscriberResponse,),
                'auth': [],
                'endpoint_path': '/v1/signals/{id}/subscribers',
                'operation_id': 'v1_signals_id_subscribers_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'id',
                    'key_manager_token',
                    'skip',
                    'limit',
                ],
                'required': [
                    'session_token',
                    'id',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'id':
                        (str,),
                    'key_manager_token':
                        (str,),
                    'skip':
                        (int,),
                    'limit':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'id': 'id',
                    'key_manager_token': 'keyManagerToken',
                    'skip': 'skip',
                    'limit': 'limit',
                },
                'location_map': {
                    'session_token': 'header',
                    'id': 'path',
                    'key_manager_token': 'header',
                    'skip': 'query',
                    'limit': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_signals_id_subscribers_get
        )

        def __v1_signals_id_unsubscribe_post(
            self,
            session_token,
            id,
            **kwargs
        ):
            """Unsubscribe to a Signal.  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v1_signals_id_unsubscribe_post(session_token, id, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                id (str): The id of the signal.

            Keyword Args:
                key_manager_token (str): Key Manager authentication token.. [optional]
                users ([int]): UserIds to unsubscribe (only for bulk unsubscription). [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ChannelSubscriptionResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['id'] = \
                id
            return self.call_with_http_info(**kwargs)

        self.v1_signals_id_unsubscribe_post = _Endpoint(
            settings={
                'response_type': (ChannelSubscriptionResponse,),
                'auth': [],
                'endpoint_path': '/v1/signals/{id}/unsubscribe',
                'operation_id': 'v1_signals_id_unsubscribe_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'id',
                    'key_manager_token',
                    'users',
                ],
                'required': [
                    'session_token',
                    'id',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'id':
                        (str,),
                    'key_manager_token':
                        (str,),
                    'users':
                        ([int],),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'id': 'id',
                    'key_manager_token': 'keyManagerToken',
                },
                'location_map': {
                    'session_token': 'header',
                    'id': 'path',
                    'key_manager_token': 'header',
                    'users': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_signals_id_unsubscribe_post
        )

        def __v1_signals_id_update_post(
            self,
            session_token,
            id,
            signal,
            **kwargs
        ):
            """Update a signal.  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v1_signals_id_update_post(session_token, id, signal, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                id (str): The id of the signal.
                signal (BaseSignal): Signal definition.

            Keyword Args:
                key_manager_token (str): Key Manager authentication token.. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Signal
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['id'] = \
                id
            kwargs['signal'] = \
                signal
            return self.call_with_http_info(**kwargs)

        self.v1_signals_id_update_post = _Endpoint(
            settings={
                'response_type': (Signal,),
                'auth': [],
                'endpoint_path': '/v1/signals/{id}/update',
                'operation_id': 'v1_signals_id_update_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'id',
                    'signal',
                    'key_manager_token',
                ],
                'required': [
                    'session_token',
                    'id',
                    'signal',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'id':
                        (str,),
                    'signal':
                        (BaseSignal,),
                    'key_manager_token':
                        (str,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'id': 'id',
                    'key_manager_token': 'keyManagerToken',
                },
                'location_map': {
                    'session_token': 'header',
                    'id': 'path',
                    'signal': 'body',
                    'key_manager_token': 'header',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_signals_id_update_post
        )

        def __v1_signals_list_get(
            self,
            session_token,
            **kwargs
        ):
            """List signals for the requesting user. This includes signals that the user has created and public signals to which they subscribed.   # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v1_signals_list_get(session_token, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.

            Keyword Args:
                key_manager_token (str): Key Manager authentication token.. [optional]
                skip (int): No. of signals to skip. . [optional]
                limit (int): Max no. of signals to return. If no value is provided, 50 is the default. The maximum supported value is 500. . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SignalList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            return self.call_with_http_info(**kwargs)

        self.v1_signals_list_get = _Endpoint(
            settings={
                'response_type': (SignalList,),
                'auth': [],
                'endpoint_path': '/v1/signals/list',
                'operation_id': 'v1_signals_list_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'key_manager_token',
                    'skip',
                    'limit',
                ],
                'required': [
                    'session_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'key_manager_token':
                        (str,),
                    'skip':
                        (int,),
                    'limit':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'key_manager_token': 'keyManagerToken',
                    'skip': 'skip',
                    'limit': 'limit',
                },
                'location_map': {
                    'session_token': 'header',
                    'key_manager_token': 'header',
                    'skip': 'query',
                    'limit': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_signals_list_get
        )



from abc import ABC, abstractmethod
from typing import TypeVar, Generic

from symphony.bdk.gen.agent_model.v4_initiator import V4Initiator

C = TypeVar('C')  # Context type
E = TypeVar('E')  # Event type


class ActivityContext(Generic[E]):
    """
    Base class for any activity context holder.
    """

    def __init__(self, initiator: V4Initiator, source_event: E):
        """

        :param initiator: The activity initiator  (i.e. the Symphony user that triggered an event in the chat)
        :param source_event: The activity source real-time event.
        """
        self._initiator = initiator
        self._source_event = source_event

    @property
    def initiator(self):
        """

        :return: The activity initiator  (i.e. the Symphony user that triggered an event in the chat)
        """
        return self._initiator

    @property
    def source_event(self):
        """

        :return: The activity source real-time event.
        """
        return self._source_event


class AbstractActivity(ABC, Generic[C]):
    """
    Base abstract class for activities provided by the BDK. Provides a generic flow to process an incoming chat event.
    """

    @abstractmethod
    def matches(self, context: C) -> bool:
        """Matches the :py:class:`ActivityContext` to decide whether an :py:class:`AbstractActivity` can be executed
        or not.

        :param context: an instance of :py:class:`ActivityContext`
        :return: True if activity has to be triggered, False otherwise.
        """

    @abstractmethod
    async def on_activity(self, context: C):
        """Contains the activity business logic. Executed only if the `matches` method returned a true value.

        :param context: an instance of :py:class:`ActivityContext`
        """

    def before_matcher(self, context: C):
        """ This callback can be used to prepare :py:class:`ActivityContext` before actually processing the `matches`
        method.

        :param context: an instance of :py:class:`ActivityContext`
        """



"""
    Pod API

    This document refers to Symphony API calls that do not need encryption or decryption of content.  - sessionToken can be obtained by calling the authenticationAPI on the symphony back end and the key manager respectively. Refer to the methods described in authenticatorAPI.yaml. - Actions are defined to be atomic, ie will succeed in their entirety or fail and have changed nothing. - If it returns a 40X status then it will have made no change to the system even if ome subset of the request would have succeeded. - If this contract cannot be met for any reason then this is an error and the response code will be 50X.   # noqa: E501

    The version of the OpenAPI document: 20.12.1
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from symphony.bdk.gen.api_client import ApiClient, Endpoint as _Endpoint
from symphony.bdk.gen.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from symphony.bdk.gen.pod_model.company_cert import CompanyCert
from symphony.bdk.gen.pod_model.company_cert_attributes import CompanyCertAttributes
from symphony.bdk.gen.pod_model.company_cert_detail import CompanyCertDetail
from symphony.bdk.gen.pod_model.company_cert_info_list import CompanyCertInfoList
from symphony.bdk.gen.pod_model.company_cert_type_list import CompanyCertTypeList
from symphony.bdk.gen.pod_model.error import Error
from symphony.bdk.gen.pod_model.string_id import StringId
from symphony.bdk.gen.pod_model.success_response import SuccessResponse


class SecurityApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __v1_companycert_delete_post(
            self,
            session_token,
            finger_print,
            **kwargs
        ):
            """Delete a company certificate  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_companycert_delete_post(session_token, finger_print, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                finger_print (StringId):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['finger_print'] = \
                finger_print
            return self.call_with_http_info(**kwargs)

        self.v1_companycert_delete_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/companycert/delete',
                'operation_id': 'v1_companycert_delete_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'finger_print',
                ],
                'required': [
                    'session_token',
                    'finger_print',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'finger_print':
                        (StringId,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                },
                'location_map': {
                    'session_token': 'header',
                    'finger_print': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_companycert_delete_post
        )

        def __v1_companycert_finger_print_get_get(
            self,
            finger_print,
            session_token,
            **kwargs
        ):
            """Get the details of a company certificate  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_companycert_finger_print_get_get(finger_print, session_token, async_req=True)
            >>> result = thread.get()

            Args:
                finger_print (str): Certificate fingerPrint (ID)
                session_token (str): Session authentication token.

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                CompanyCertDetail
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['finger_print'] = \
                finger_print
            kwargs['session_token'] = \
                session_token
            return self.call_with_http_info(**kwargs)

        self.v1_companycert_finger_print_get_get = _Endpoint(
            settings={
                'response_type': (CompanyCertDetail,),
                'auth': [],
                'endpoint_path': '/v1/companycert/{fingerPrint}/get',
                'operation_id': 'v1_companycert_finger_print_get_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'finger_print',
                    'session_token',
                ],
                'required': [
                    'finger_print',
                    'session_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'finger_print':
                        (str,),
                    'session_token':
                        (str,),
                },
                'attribute_map': {
                    'finger_print': 'fingerPrint',
                    'session_token': 'sessionToken',
                },
                'location_map': {
                    'finger_print': 'path',
                    'session_token': 'header',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_companycert_finger_print_get_get
        )

        def __v1_companycert_finger_print_issued_by_get(
            self,
            finger_print,
            session_token,
            **kwargs
        ):
            """Return a list of all certificates which were verified to the cert whose fingerprint is passed.   # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_companycert_finger_print_issued_by_get(finger_print, session_token, async_req=True)
            >>> result = thread.get()

            Args:
                finger_print (str): Certificate fingerPrint (ID)
                session_token (str): Session authentication token.

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                CompanyCertInfoList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['finger_print'] = \
                finger_print
            kwargs['session_token'] = \
                session_token
            return self.call_with_http_info(**kwargs)

        self.v1_companycert_finger_print_issued_by_get = _Endpoint(
            settings={
                'response_type': (CompanyCertInfoList,),
                'auth': [],
                'endpoint_path': '/v1/companycert/{fingerPrint}/issuedBy',
                'operation_id': 'v1_companycert_finger_print_issued_by_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'finger_print',
                    'session_token',
                ],
                'required': [
                    'finger_print',
                    'session_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'finger_print':
                        (str,),
                    'session_token':
                        (str,),
                },
                'attribute_map': {
                    'finger_print': 'fingerPrint',
                    'session_token': 'sessionToken',
                },
                'location_map': {
                    'finger_print': 'path',
                    'session_token': 'header',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_companycert_finger_print_issued_by_get
        )

        def __v1_companycert_finger_print_update_post(
            self,
            finger_print,
            session_token,
            cert_attributes,
            **kwargs
        ):
            """Update a company certificate  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_companycert_finger_print_update_post(finger_print, session_token, cert_attributes, async_req=True)
            >>> result = thread.get()

            Args:
                finger_print (str): Certificate fingerPrint (ID)
                session_token (str): Session authentication token.
                cert_attributes (CompanyCertAttributes):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                SuccessResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['finger_print'] = \
                finger_print
            kwargs['session_token'] = \
                session_token
            kwargs['cert_attributes'] = \
                cert_attributes
            return self.call_with_http_info(**kwargs)

        self.v1_companycert_finger_print_update_post = _Endpoint(
            settings={
                'response_type': (SuccessResponse,),
                'auth': [],
                'endpoint_path': '/v1/companycert/{fingerPrint}/update',
                'operation_id': 'v1_companycert_finger_print_update_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'finger_print',
                    'session_token',
                    'cert_attributes',
                ],
                'required': [
                    'finger_print',
                    'session_token',
                    'cert_attributes',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'finger_print':
                        (str,),
                    'session_token':
                        (str,),
                    'cert_attributes':
                        (CompanyCertAttributes,),
                },
                'attribute_map': {
                    'finger_print': 'fingerPrint',
                    'session_token': 'sessionToken',
                },
                'location_map': {
                    'finger_print': 'path',
                    'session_token': 'header',
                    'cert_attributes': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_companycert_finger_print_update_post
        )

        def __v1_companycert_list_get(
            self,
            session_token,
            **kwargs
        ):
            """List all trusted certs  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_companycert_list_get(session_token, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.

            Keyword Args:
                skip (int): Pagination start. [optional]
                limit (int): Row limit. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                CompanyCertInfoList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            return self.call_with_http_info(**kwargs)

        self.v1_companycert_list_get = _Endpoint(
            settings={
                'response_type': (CompanyCertInfoList,),
                'auth': [],
                'endpoint_path': '/v1/companycert/list',
                'operation_id': 'v1_companycert_list_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'skip',
                    'limit',
                ],
                'required': [
                    'session_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'skip':
                        (int,),
                    'limit':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'skip': 'skip',
                    'limit': 'limit',
                },
                'location_map': {
                    'session_token': 'header',
                    'skip': 'query',
                    'limit': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_companycert_list_get
        )

        def __v1_companycert_podmanaged_list_get(
            self,
            session_token,
            **kwargs
        ):
            """List all trusted certs  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_companycert_podmanaged_list_get(session_token, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.

            Keyword Args:
                skip (int): Pagination start. [optional]
                limit (int): Row limit. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                CompanyCertInfoList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            return self.call_with_http_info(**kwargs)

        self.v1_companycert_podmanaged_list_get = _Endpoint(
            settings={
                'response_type': (CompanyCertInfoList,),
                'auth': [],
                'endpoint_path': '/v1/companycert/podmanaged/list',
                'operation_id': 'v1_companycert_podmanaged_list_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'skip',
                    'limit',
                ],
                'required': [
                    'session_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'skip':
                        (int,),
                    'limit':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'skip': 'skip',
                    'limit': 'limit',
                },
                'location_map': {
                    'session_token': 'header',
                    'skip': 'query',
                    'limit': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v1_companycert_podmanaged_list_get
        )

        def __v1_companycert_type_list_post(
            self,
            session_token,
            type_id_list,
            **kwargs
        ):
            """List all certs of the given types  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v1_companycert_type_list_post(session_token, type_id_list, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                type_id_list (CompanyCertTypeList): Certificate type list

            Keyword Args:
                skip (int): Pagination start. [optional]
                limit (int): Row limit. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                CompanyCertInfoList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['type_id_list'] = \
                type_id_list
            return self.call_with_http_info(**kwargs)

        self.v1_companycert_type_list_post = _Endpoint(
            settings={
                'response_type': (CompanyCertInfoList,),
                'auth': [],
                'endpoint_path': '/v1/companycert/type/list',
                'operation_id': 'v1_companycert_type_list_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'type_id_list',
                    'skip',
                    'limit',
                ],
                'required': [
                    'session_token',
                    'type_id_list',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'type_id_list':
                        (CompanyCertTypeList,),
                    'skip':
                        (int,),
                    'limit':
                        (int,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'skip': 'skip',
                    'limit': 'limit',
                },
                'location_map': {
                    'session_token': 'header',
                    'type_id_list': 'body',
                    'skip': 'query',
                    'limit': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v1_companycert_type_list_post
        )

        def __v2_companycert_create_post(
            self,
            session_token,
            cert,
            **kwargs
        ):
            """Create a company trusted or untrusted certificate. Different from V1 in that we reject expired certificates.  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = pod_api.v2_companycert_create_post(session_token, cert, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                cert (CompanyCert):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                CompanyCertDetail
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['cert'] = \
                cert
            return self.call_with_http_info(**kwargs)

        self.v2_companycert_create_post = _Endpoint(
            settings={
                'response_type': (CompanyCertDetail,),
                'auth': [],
                'endpoint_path': '/v2/companycert/create',
                'operation_id': 'v2_companycert_create_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'cert',
                ],
                'required': [
                    'session_token',
                    'cert',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'cert':
                        (CompanyCert,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                },
                'location_map': {
                    'session_token': 'header',
                    'cert': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__v2_companycert_create_post
        )



"""This module contains a set of util functions that can be used on incoming V4Message such as:
extracting entities like Mentions, Hashtags, Cashtags, Emojis and extracting text content from presentationML
contained in the message.
"""
import json
import html

from enum import Enum
from json import JSONDecodeError
from typing import Dict

from defusedxml.ElementTree import fromstring, tostring, ParseError

from symphony.bdk.core.service.exception import MessageParserError
from symphony.bdk.gen.agent_model.v4_message import V4Message


def get_text_content_from_message(message: V4Message) -> str:
    """Get text content from PresentationML on incoming messages

    :param message: message containing the PresentationML to be parsed
    :return: the message text content extracted from the given PresentationML
    """
    try:
        presentation_ml = html.unescape(message.message)
        return tostring(fromstring(presentation_ml), method="text").decode().strip()
    except ParseError as exc:
        raise MessageParserError("Unable to parse the PresentationML, it is not in the correct format.") from exc


def get_mentions(message: V4Message) -> [int]:
    """ Parse data inside an incoming message and returns a list containing the user ids corresponding
    to the users mentioned

    :param message: incoming V4 message to be parsed
    :return: list of users ids that has been mentioned inside the message
    """
    mentions_list = _get_tags(message, _EntityTypeEnum.MENTION)
    return [int(user_id) for user_id in mentions_list]


def get_hashtags(message: V4Message) -> [str]:
    """ Parse data inside an incoming message and returns a list containing the text of the hashtags found

    :param message: message incoming V4 message to be parsed
    :return: list of hashtags contained in the message
    """
    return _get_tags(message, _EntityTypeEnum.HASHTAG)


def get_cashtags(message: V4Message) -> [str]:
    """ Parse data inside an incoming message and returns a list containing the text of the cashtags found

    :param message: message incoming V4 message to be parsed
    :return: list of cashtags contained in the message
    """
    return _get_tags(message, _EntityTypeEnum.CASHTAG)


def get_emojis(message: V4Message) -> Dict[str, str]:
    """ Parse data inside an incoming message and returns a map containing the list of emojis found.
    Key of the map are the annotation used to identify the emoji and the values are the their unicode.

    :param message: message incoming V4 message to be parsed
    :return: list of cashtags contained in the message
    """
    json_data = _parse_json_data(message.data)
    emojis_list = {}
    for item in json_data.values():
        try:
            if item["type"] == _EntityTypeEnum.EMOJI.value:
                emojis_list[item["data"]["annotation"]] = item["data"]["unicode"]
        except KeyError:
            pass
    return emojis_list


def _get_tags(message, entity_type):
    json_data = _parse_json_data(message.data)
    tags_list = []
    for item in json_data.values():
        try:
            if item["type"] == entity_type.value:
                tags_list.append(item["id"][0]["value"])
        except KeyError:
            pass
    return tags_list


def _parse_json_data(json_data):
    try:
        return json.loads(json_data)
    except JSONDecodeError as exc:
        raise MessageParserError("Failed to extract payload from message data") from exc


class _EntityTypeEnum(Enum):
    HASHTAG = "org.symphonyoss.taxonomy"
    CASHTAG = "org.symphonyoss.fin.security"
    MENTION = "com.symphony.user.mention"
    EMOJI = "com.symphony.emoji"



"""Module containing the ApiClientFactory class.
"""
import logging
import sys
from importlib.metadata import distribution, PackageNotFoundError
from ssl import SSLError

import urllib3
from aiohttp.hdrs import USER_AGENT

from symphony.bdk.core.client.trace_id import add_x_trace_id, X_TRACE_ID
from symphony.bdk.gen.api_client import ApiClient
from symphony.bdk.gen.configuration import Configuration

KEY_AUTH = "/keyauth"
SESSION_AUTH = "/sessionauth"
AGENT = "/agent"
RELAY = "/relay"
POD = "/pod"
LOGIN = "/login"

logger = logging.getLogger(__name__)


class ApiClientFactory:
    """Factory responsible for creating ApiClient instances for each main Symphony's components.
    """

    def __init__(self, config):
        self._config = config
        self._login_client = self._get_api_client(self._config.pod, LOGIN)
        self._pod_client = self._get_api_client(self._config.pod, POD)
        self._relay_client = self._get_api_client(self._config.key_manager, RELAY)
        self._agent_client = self._get_api_client(self._config.agent, AGENT)
        self._session_auth_client = self._get_api_client_with_client_cert(self._config.session_auth, SESSION_AUTH,
                                                                          self._config.bot.certificate.path)
        self._key_auth_client = self._get_api_client_with_client_cert(self._config.key_manager, KEY_AUTH,
                                                                      self._config.bot.certificate.path)
        self._app_session_auth_client = self._get_api_client_with_client_cert(self._config.session_auth, SESSION_AUTH,
                                                                              self._config.app.certificate.path)

    def get_login_client(self) -> ApiClient:
        """Returns a fully initialized ApiClient for Login API.

        :return: an ApiClient instance for Login API.
        """
        return self._login_client

    def get_pod_client(self) -> ApiClient:
        """Returns a fully initialized ApiClient for Pod API.

        :return: an ApiClient instance for Pod API.
        """
        return self._pod_client

    def get_relay_client(self) -> ApiClient:
        """Returns a fully initialized ApiClient for Key Manager API.

        :return: an ApiClient instance for Key Manager API.
        """
        return self._relay_client

    def get_session_auth_client(self) -> ApiClient:
        """Returns a fully initialized ApiClient for Session Auth Api for bot certificate authentication.

        :return: an ApiClient instance for Session Auth Api for bot certificate authentication.
        """
        return self._session_auth_client

    def get_key_auth_client(self) -> ApiClient:
        """Returns a fully initialized ApiClient for Key Auth Api for bot certificate authentication.

        :return: an ApiClient instance for Key Auth Api for bot certificate authentication.
        """
        return self._key_auth_client

    def get_app_session_auth_client(self) -> ApiClient:
        """Returns a fully initialized ApiClient for Session Auth API for certificate OBO authentication.

        :return: an ApiClient instance for Session Auth API.
        """
        return self._app_session_auth_client

    def get_agent_client(self) -> ApiClient:
        """Returns a fully initialized ApiClient for Agent API.

        :return: an ApiClient instance for Agent API.
        """
        return self._agent_client

    async def close_clients(self):
        """
        Close all the existing api clients created by the api client factory.
        """
        await self._login_client.close()
        await self._relay_client.close()
        await self._pod_client.close()
        await self._agent_client.close()
        await self._session_auth_client.close()
        await self._key_auth_client.close()
        await self._app_session_auth_client.close()

    def _get_api_client(self, server_config, context) -> ApiClient:
        configuration = self._get_client_config(context, server_config)
        return ApiClientFactory._get_api_client_from_config(configuration, server_config)

    def _get_api_client_with_client_cert(self, server_config, context, certificate_path) -> ApiClient:
        configuration = self._get_client_config(context, server_config)
        configuration.cert_file = certificate_path

        return ApiClientFactory._get_api_client_from_config(configuration, server_config)

    def _get_client_config(self, context, server_config):
        configuration = Configuration(host=(server_config.get_base_path() + context), discard_unknown_keys=True)
        configuration.verify_ssl = True
        configuration.ssl_ca_cert = self._config.ssl.trust_store_path
        if server_config.proxy is not None:
            ApiClientFactory._configure_proxy(server_config, configuration)
        return configuration

    @staticmethod
    def _get_api_client_from_config(client_config, server_config):
        try:
            client = ApiClient(configuration=client_config)
            ApiClientFactory._add_headers(client, server_config)
            return client
        except SSLError as exc:
            logger.exception("SSL error when instantiating clients, please check certificates are valid")
            raise exc

    @staticmethod
    def _add_headers(client, server_config):
        default_headers = ApiClientFactory._sanitized_default_headers(server_config)

        for header_name, header_value in default_headers.items():
            client.set_default_header(header_name, header_value)
        client.user_agent = ApiClientFactory._user_agent(server_config)

        if X_TRACE_ID.lower() not in (header_name.lower() for header_name in default_headers.keys()):
            client._ApiClient__call_api = add_x_trace_id(client._ApiClient__call_api)

    @staticmethod
    def _configure_proxy(server_config, configuration):
        proxy_config = server_config.proxy
        user_agent = ApiClientFactory._user_agent(server_config)

        configuration.proxy = proxy_config.get_url()
        if proxy_config.are_credentials_defined():
            configuration.proxy_headers = urllib3.util.make_headers(proxy_basic_auth=proxy_config.get_credentials(),
                                                                    user_agent=user_agent)
        else:
            configuration.proxy_headers = urllib3.util.make_headers(user_agent=user_agent)

    @staticmethod
    def _sanitized_default_headers(server_config):
        default_headers = server_config.default_headers if server_config.default_headers is not None else {}

        # we do this because we want to handle user-agent header separately
        # for client configuration and proxy header
        return {header_key: default_headers[header_key] for header_key in default_headers
                if header_key.lower() != USER_AGENT.lower()}

    @staticmethod
    def _user_agent(server_config):
        default_headers = server_config.default_headers if server_config.default_headers is not None else {}

        for header_key, header_value in default_headers.items():
            if header_key.lower() == USER_AGENT.lower():
                return header_value

        return ApiClientFactory._default_user_agent()

    @staticmethod
    def _default_user_agent():
        return f"Symphony-BDK-Python/{ApiClientFactory._bdk_version()} Python/{ApiClientFactory._python_version()}"

    @staticmethod
    def _bdk_version():
        try:
            return distribution("sym_api_client_python").version
        except PackageNotFoundError:
            # the above won't work when bdk is not installed as a pypi package,
            # e.g. for scripts in the examples folder
            return "2.0"

    @staticmethod
    def _python_version():
        return f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"



"""Module to help with jwt handling.
"""
import datetime

import jwt
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from cryptography.x509 import load_pem_x509_certificate

from symphony.bdk.core.auth.exception import AuthInitializationError
from symphony.bdk.core.config.model.bdk_rsa_key_config import BdkRsaKeyConfig

JWT_ENCRYPTION_ALGORITHM = "RS512"

DEFAULT_EXPIRATION_SECONDS = (5 * 50) - 10


def create_signed_jwt(private_key_config: BdkRsaKeyConfig, username: str, expiration: int = None) -> str:
    """Creates a JWT with the provided user name and expiration date, signed with the provided private key.

    :param private_key_config:  The private key configuration for a service account or an extension app.
    :param username:            The username of the user to authenticate.
    :param expiration:          Expiration of the authentication request in seconds.
                                By default the signed jwt will be valid in maximum 290 seconds (4min 50s)
                                which is the maximum expiration accepted by the Symphony backend.

    :return: a signed JWT for a specific user or an extension app.
    """
    expiration = expiration if expiration is not None else int(
        datetime.datetime.now(datetime.timezone.utc).timestamp() + DEFAULT_EXPIRATION_SECONDS)
    payload = {
        "sub": username,
        "exp": expiration
    }
    return create_signed_jwt_with_claims(private_key_config.get_private_key_content(), payload)


def create_signed_jwt_with_claims(private_key: str, payload: dict) -> str:
    """Creates a JWT with the payload signed with the provided private key content.
    For testing purposes only.

    :param private_key: the privste key content in string format.
    :param payload: the payload (aka claims) of the JWT in dict format.
    :return: a signed JWT
    """
    return jwt.encode(payload, private_key, algorithm=JWT_ENCRYPTION_ALGORITHM)


def validate_jwt(jwt_token: str, certificate: str, allowed_audience: str) -> dict:
    """Validate a jwt against a X509 certificate in pem format and returns the jwt claims.

    :param jwt_token: the token to be validated
    :param certificate: the X509 certificate in pem format to be used for jwt validation
    :param allowed_audience: the expected value in "aud" claim. If it doesn't match jwt will be rejected
    :return: a dictionary containing the jwt claims
    :raise AuthInitializationError: If the certificate or jwt are invalid.
    """
    try:
        return jwt.decode(jwt_token, _parse_public_key_from_x509_cert(certificate),
                          algorithms=[JWT_ENCRYPTION_ALGORITHM], audience=allowed_audience)
    except (jwt.DecodeError, jwt.ExpiredSignatureError) as exc:
        raise AuthInitializationError("Unable to validate the jwt") from exc


def _parse_public_key_from_x509_cert(certificate: str) -> str:
    """Returns the public key in PEM format given a X509 certificate content in PEM format

    :param certificate: the X509 certificate in PEM format
    :return: the public key associated to the certificate in PEM format
    """
    try:
        public_key = load_pem_x509_certificate(certificate.encode()).public_key()
        return public_key.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo).decode()
    except ValueError as exc:
        raise AuthInitializationError("Unable to parse the certificate. Check certificate format.") from exc



import sys
from datetime import timedelta


class BdkRetryConfig:
    """Class containing the retry configuration"""
    INFINITE_MAX_ATTEMPTS = -1
    DEFAULT_MAX_ATTEMPTS = 10
    DEFAULT_INITIAL_INTERVAL = 500
    DEFAULT_MULTIPLIER = 2
    DEFAULT_MAX_INTERVAL = 5 * 60 * 1000

    def __init__(self, config=None):
        if config is None:
            config = {}

        self.max_attempts = config.get("maxAttempts", self.DEFAULT_MAX_ATTEMPTS)
        if self.max_attempts < 0:
            self.max_attempts = sys.maxsize
        self.initial_interval = timedelta(milliseconds=config.get("initialIntervalMillis",
                                                                  self.DEFAULT_INITIAL_INTERVAL))
        self.multiplier = config.get("multiplier", self.DEFAULT_MULTIPLIER)
        if self.multiplier < 1:
            self.multiplier = self.DEFAULT_MULTIPLIER
        self.max_interval = timedelta(milliseconds=config.get("maxIntervalMillis", self.DEFAULT_MAX_INTERVAL))



from symphony.bdk.core.auth.auth_session import AuthSession
from symphony.bdk.core.config.model.bdk_retry_config import BdkRetryConfig
from symphony.bdk.core.retry import retry
from symphony.bdk.core.service.connection.model.connection_status import ConnectionStatus
from symphony.bdk.gen.pod_api.connection_api import ConnectionApi
from symphony.bdk.gen.pod_model.user_connection import UserConnection
from symphony.bdk.gen.pod_model.user_connection_request import UserConnectionRequest


class OboConnectionService:
    """Class exposing OBO-enabled endpoints for connection management.

    This service is used for retrieving the connection status between the OBO user and a specified user or several
    other internal or external users in the pod, and perform some actions related to the connection status like:

    * Send a connection request to an user
    * Accept a connection request from a user
    * Reject a connection request from a user
    * Remove a connection with a user
    """

    def __init__(self, connection_api: ConnectionApi, auth_session: AuthSession, retry_config: BdkRetryConfig):
        self._connection_api = connection_api
        self._auth_session = auth_session
        self._retry_config = retry_config

    @retry
    async def get_connection(self, user_id: int) -> UserConnection:
        """
        Get connection status, i.e. check if the calling user is connected to the specified user.
        See: `Get Connection <https://developers.symphony.com/restapi/reference#get-connection>`_

        :param user_id: The id of the user with whom the caller want to check.

        :return: Connection status with the specified user.

        """
        params = {
            'user_id': str(user_id),
            'session_token': await self._auth_session.session_token
        }
        return await self._connection_api.v1_connection_user_user_id_info_get(**params)

    @retry
    async def list_connections(
            self,
            status: ConnectionStatus = ConnectionStatus.ALL,
            user_ids: [int] = None
    ) -> [UserConnection]:
        """
        List all connection statuses of the requesting user with external or specified users.
        See: `List Connections <https://developers.symphony.com/restapi/reference#list-connections>`_

        :param status:      Filter the connection list based on the connection status.
                            The connection status can only be pending_incoming, pending_outgoing,
                            accepted, rejected, or all.
                            If you do not specify a status, all connections will be returned.
        :param user_ids:    List of user ids which are used to restrict the list of results.
                            This can be used to return connections with internal users;
                            although, by default, this endpoint does not list implicit connections with internal users.

        :return: List of connection statuses with the specified users and status.

        """
        params = {
            'status': status.value,
            'session_token': await self._auth_session.session_token
        }
        if user_ids is not None:
            params['user_ids'] = ','.join(map(str, user_ids))

        user_connection_list = await self._connection_api.v1_connection_list_get(**params)
        return user_connection_list.value

    @retry
    async def create_connection(self, user_id: int) -> UserConnection:
        """
        Sends a connection request to another user.
        See: `Create Connection <https://developers.symphony.com/restapi/reference#create-connection>`_

        :param user_id: The id of the user with whom the caller want to connect.

        :return: Connection status with the specified user.

        """
        user_connection_request = UserConnectionRequest(user_id=user_id)
        params = {
            'connection_request': user_connection_request,
            'session_token': await self._auth_session.session_token
        }
        return await self._connection_api.v1_connection_create_post(**params)

    @retry
    async def accept_connection(self, user_id: int) -> UserConnection:
        """
        Accept the connection request from a requesting user.
        See: `Accept Connection <https://developers.symphony.com/restapi/reference#accepted-connection>`_

        :param user_id: The id of the user who requested to connect with the caller.

        :return: Connection status with the requesting user.

        """
        user_connection_request = UserConnectionRequest(user_id=user_id)
        params = {
            'connection_request': user_connection_request,
            'session_token': await self._auth_session.session_token
        }
        return await self._connection_api.v1_connection_accept_post(**params)

    @retry
    async def reject_connection(self, user_id: int) -> UserConnection:
        """
        Reject the connection request from a requesting user.
        See: `Reject Connection <https://developers.symphony.com/restapi/reference#reject-connection>`_

        :param user_id: The id of the user who requested to connect with the caller.

        :return: Connection status with the requesting user.

        """
        user_connection_request = UserConnectionRequest(user_id=user_id)
        params = {
            'connection_request': user_connection_request,
            'session_token': await self._auth_session.session_token
        }
        return await self._connection_api.v1_connection_reject_post(**params)

    @retry
    async def remove_connection(self, user_id: int) -> None:
        """
        Removes a connection with a user.
        See: `Remove Connection <https://developers.symphony.com/restapi/reference#remove-connection>`_

        :param user_id: The id of the user with whom we want to remove the connection.

        """
        params = {
            'uid': user_id,
            'session_token': await self._auth_session.session_token
        }
        await self._connection_api.v1_connection_user_uid_remove_post(**params)


class ConnectionService(OboConnectionService):
    """Service class for managing the connections between users

    This service is used for retrieving the connection status between the calling user and a specified user or several
    other internal or external users in the pod, and perform some actions related to the connection status like:

    * Send a connection request to an user
    * Accept a connection request from a user
    * Reject a connection request from a user
    * Remove a connection with a user
    """



"""This module handles pre-processing for an outgoing message in order to have a valid messageML format.
"""
import re


def escape_special_chars(raw_text: str) -> str:
    """ Replace all special characters placed within the messageML that must be HTML-escaped
    to have a valid MessageML format.

    :param raw_text: text to be parsed
    :return: text in a valid messageML format
    """
    current_parsed_index = 0
    parsed_text = ""
    matches = re.finditer(_pattern(), raw_text)
    for match in matches:
        parsed_text += raw_text[current_parsed_index:match.start()] + _replacement(match.group(0))
        current_parsed_index = match.end()
    return parsed_text + raw_text[current_parsed_index:]


def _pattern():
    return "|".join(_special_chars_dict)


def _replacement(match):
    return _special_chars_dict.get(match, _special_chars_dict.get("\\" + match))


_special_chars_dict = {
    "&": "&amp;",
    "<": "&lt;",
    ">": "&gt;",
    "'": "&apos;",
    "\"": "&quot;",
    "#": "&#35;",
    "\\$": "&#36;",
    "%": "&#37;",
    "\\(": "&#40;",
    "\\)": "&#41;",
    "\\*": "&#42;",
    "\\.": "&#46;",
    ";": "&#59;",
    "=": "&#61;",
    "\\[": "&#91;",
    "\\\\": "&#92;",
    "\\]": "&#93;",
    "`": "&#96;",
    "\\{": "&#123;",
    "\\}": "&#125;"
}



import logging

from symphony.bdk.core.activity.api import AbstractActivity, ActivityContext
from symphony.bdk.gen.agent_model.v4_initiator import V4Initiator
from symphony.bdk.gen.agent_model.v4_user_joined_room import V4UserJoinedRoom

logger = logging.getLogger(__name__)


class UserJoinedRoomContext(ActivityContext[V4UserJoinedRoom]):
    """
    Default implementation of the :py:class`ActivityContext` handled by the :py:class:`UserJoinedRoomActivity`
    """

    def __init__(self, initiator: V4Initiator, source_event: V4UserJoinedRoom):
        self._stream_id = source_event.stream.stream_id
        self._affected_user_id = source_event.affected_user.user_id
        super().__init__(initiator, source_event)

    @property
    def stream_id(self) -> str:
        return self._stream_id

    @property
    def affected_user_id(self) -> int:
        return self._affected_user_id


class UserJoinedRoomActivity(AbstractActivity[UserJoinedRoomContext]):
    """
    A form reply activity corresponds to a User joining a room .
    """

    def matches(self, context: UserJoinedRoomContext) -> bool:
        pass

    async def on_activity(self, context: UserJoinedRoomContext):
        pass



"""
    Agent API

    This document refers to Symphony API calls to send and receive messages and content. They need the on-premise Agent installed to perform decryption/encryption of content.  - sessionToken and keyManagerToken can be obtained by calling the authenticationAPI on the symphony back end and the key manager respectively. Refer to the methods described in authenticatorAPI.yaml. - Actions are defined to be atomic, ie will succeed in their entirety or fail and have changed nothing. - If it returns a 40X status then it will have sent no message to any stream even if a request to aome subset of the requested streams would have succeeded. - If this contract cannot be met for any reason then this is an error and the response code will be 50X. - MessageML is a markup language for messages. See reference here: https://rest-api.symphony.com/docs/messagemlv2 - **Real Time Events**: The following events are returned when reading from a real time messages and events stream (\"datafeed\"). These events will be returned for datafeeds created with the v5 endpoints. To know more about the endpoints, refer to Create Messages/Events Stream and Read Messages/Events Stream. Unless otherwise specified, all events were added in 1.46.   # noqa: E501

    The version of the OpenAPI document: 20.12.0
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from symphony.bdk.gen.api_client import ApiClient, Endpoint as _Endpoint
from symphony.bdk.gen.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from symphony.bdk.gen.agent_model.ack_id import AckId
from symphony.bdk.gen.agent_model.datafeed import Datafeed
from symphony.bdk.gen.agent_model.v2_error import V2Error
from symphony.bdk.gen.agent_model.v4_event_list import V4EventList
from symphony.bdk.gen.agent_model.v5_datafeed import V5Datafeed
from symphony.bdk.gen.agent_model.v5_datafeed_create_body import V5DatafeedCreateBody
from symphony.bdk.gen.agent_model.v5_event_list import V5EventList


class DatafeedApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __create_datafeed(
            self,
            session_token,
            key_manager_token,
            **kwargs
        ):
            """Create a new real time messages / events stream (\"datafeed\").  # noqa: E501

            _Available on Agent 2.57.0 and above._  The datafeed provides messages and events from all conversations that the user is in. The types of events surfaced in the datafeed can be found in the Real Time Events list. (see definition on top of the file)  Returns the ID of the datafeed that has just been created. This ID should then be used as input to the Read Messages/Events Stream v4 endpoint.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.create_datafeed(session_token, key_manager_token, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                key_manager_token (str): Key Manager authentication token.

            Keyword Args:
                body (V5DatafeedCreateBody): [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                V5Datafeed
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['key_manager_token'] = \
                key_manager_token
            return self.call_with_http_info(**kwargs)

        self.create_datafeed = _Endpoint(
            settings={
                'response_type': (V5Datafeed,),
                'auth': [],
                'endpoint_path': '/v5/datafeeds',
                'operation_id': 'create_datafeed',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'key_manager_token',
                    'body',
                ],
                'required': [
                    'session_token',
                    'key_manager_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'key_manager_token':
                        (str,),
                    'body':
                        (V5DatafeedCreateBody,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'key_manager_token': 'keyManagerToken',
                },
                'location_map': {
                    'session_token': 'header',
                    'key_manager_token': 'header',
                    'body': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__create_datafeed
        )

        def __delete_datafeed(
            self,
            datafeed_id,
            session_token,
            key_manager_token,
            **kwargs
        ):
            """Delete the specified real time message / event stream (\"datafeed\").  # noqa: E501

            _Available on Agent 2.57.0 and above._  The datafeed provides messages and events from all conversations that the user is in. The types of events surfaced in the datafeed can be found in the Real Time Events list. (see definition on top of the file)  Delete the specified datafeed.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.delete_datafeed(datafeed_id, session_token, key_manager_token, async_req=True)
            >>> result = thread.get()

            Args:
                datafeed_id (str): ID of the datafeed
                session_token (str): Session authentication token.
                key_manager_token (str): Key Manager authentication token.

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                V2Error
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['datafeed_id'] = \
                datafeed_id
            kwargs['session_token'] = \
                session_token
            kwargs['key_manager_token'] = \
                key_manager_token
            return self.call_with_http_info(**kwargs)

        self.delete_datafeed = _Endpoint(
            settings={
                'response_type': (V2Error,),
                'auth': [],
                'endpoint_path': '/v5/datafeeds/{datafeedId}',
                'operation_id': 'delete_datafeed',
                'http_method': 'DELETE',
                'servers': None,
            },
            params_map={
                'all': [
                    'datafeed_id',
                    'session_token',
                    'key_manager_token',
                ],
                'required': [
                    'datafeed_id',
                    'session_token',
                    'key_manager_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'datafeed_id':
                        (str,),
                    'session_token':
                        (str,),
                    'key_manager_token':
                        (str,),
                },
                'attribute_map': {
                    'datafeed_id': 'datafeedId',
                    'session_token': 'sessionToken',
                    'key_manager_token': 'keyManagerToken',
                },
                'location_map': {
                    'datafeed_id': 'path',
                    'session_token': 'header',
                    'key_manager_token': 'header',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    '*/*'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__delete_datafeed
        )

        def __list_datafeed(
            self,
            session_token,
            key_manager_token,
            **kwargs
        ):
            """Read list of real time messages / events stream (\"datafeed\").  # noqa: E501

            _Available on Agent 2.57.0 and above._  The datafeed provides messages and events from all conversations that the user is in. The types of events surfaced in the datafeed can be found in the [Real Time Events](./docs/real-time-events.md) list.  Returns the list of the datafeeds for the user. Any datafeed ID of the list can then be used as input to the Read Messages/Events Stream v4 endpoint.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.list_datafeed(session_token, key_manager_token, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                key_manager_token (str): Key Manager authentication token.

            Keyword Args:
                tag (str): A unique identifier to ensure uniqueness of the datafeed. Used to restrict search.. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                [V5Datafeed]
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['key_manager_token'] = \
                key_manager_token
            return self.call_with_http_info(**kwargs)

        self.list_datafeed = _Endpoint(
            settings={
                'response_type': ([V5Datafeed],),
                'auth': [],
                'endpoint_path': '/v5/datafeeds',
                'operation_id': 'list_datafeed',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'key_manager_token',
                    'tag',
                ],
                'required': [
                    'session_token',
                    'key_manager_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                    'tag',
                ]
            },
            root_map={
                'validations': {
                    ('tag',): {
                        'max_length': 100,
                    },
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'key_manager_token':
                        (str,),
                    'tag':
                        (str,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'key_manager_token': 'keyManagerToken',
                    'tag': 'tag',
                },
                'location_map': {
                    'session_token': 'header',
                    'key_manager_token': 'header',
                    'tag': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json',
                    '200 OK'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__list_datafeed
        )

        def __read_datafeed(
            self,
            datafeed_id,
            session_token,
            key_manager_token,
            **kwargs
        ):
            """Read the specified real time message / event stream (\"datafeed\").  # noqa: E501

            _Available on Agent 2.57.0 and above._  The datafeed provides messages and events from all conversations that the user is in. The types of events surfaced in the datafeed can be found in the Real Time Events list. (see definition on top of the file)  Read the specified datafeed.  The ackId sent as parameter can be empty for the first call. In the response an ackId will be sent back and it can be used for the next call: in this way you acknowledge that you have received the events that came with that ackId; datafeed will remove the events associated with that ackId from your queue   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.read_datafeed(datafeed_id, session_token, key_manager_token, async_req=True)
            >>> result = thread.get()

            Args:
                datafeed_id (str): ID of the datafeed
                session_token (str): Session authentication token.
                key_manager_token (str): Key Manager authentication token.

            Keyword Args:
                ack_id (AckId): ackId received from last POST Base64 encoded.. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                V5EventList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['datafeed_id'] = \
                datafeed_id
            kwargs['session_token'] = \
                session_token
            kwargs['key_manager_token'] = \
                key_manager_token
            return self.call_with_http_info(**kwargs)

        self.read_datafeed = _Endpoint(
            settings={
                'response_type': (V5EventList,),
                'auth': [],
                'endpoint_path': '/v5/datafeeds/{datafeedId}/read',
                'operation_id': 'read_datafeed',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'datafeed_id',
                    'session_token',
                    'key_manager_token',
                    'ack_id',
                ],
                'required': [
                    'datafeed_id',
                    'session_token',
                    'key_manager_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'datafeed_id':
                        (str,),
                    'session_token':
                        (str,),
                    'key_manager_token':
                        (str,),
                    'ack_id':
                        (AckId,),
                },
                'attribute_map': {
                    'datafeed_id': 'datafeedId',
                    'session_token': 'sessionToken',
                    'key_manager_token': 'keyManagerToken',
                },
                'location_map': {
                    'datafeed_id': 'path',
                    'session_token': 'header',
                    'key_manager_token': 'header',
                    'ack_id': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__read_datafeed
        )

        def __v4_datafeed_create_post(
            self,
            session_token,
            key_manager_token,
            **kwargs
        ):
            """Create a new real time message event stream.  # noqa: E501

            A datafeed provides the messages in all conversations that a user is in. This also includes system messages like new users joining a chatroom.  A datafeed will expire if it isn't read before its capacity is reached.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v4_datafeed_create_post(session_token, key_manager_token, async_req=True)
            >>> result = thread.get()

            Args:
                session_token (str): Session authentication token.
                key_manager_token (str): Key Manager authentication token.

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Datafeed
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['session_token'] = \
                session_token
            kwargs['key_manager_token'] = \
                key_manager_token
            return self.call_with_http_info(**kwargs)

        self.v4_datafeed_create_post = _Endpoint(
            settings={
                'response_type': (Datafeed,),
                'auth': [],
                'endpoint_path': '/v4/datafeed/create',
                'operation_id': 'v4_datafeed_create_post',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'session_token',
                    'key_manager_token',
                ],
                'required': [
                    'session_token',
                    'key_manager_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'session_token':
                        (str,),
                    'key_manager_token':
                        (str,),
                },
                'attribute_map': {
                    'session_token': 'sessionToken',
                    'key_manager_token': 'keyManagerToken',
                },
                'location_map': {
                    'session_token': 'header',
                    'key_manager_token': 'header',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v4_datafeed_create_post
        )

        def __v4_datafeed_id_read_get(
            self,
            id,
            session_token,
            key_manager_token,
            **kwargs
        ):
            """Read a given datafeed.  # noqa: E501

            Read messages from the given datafeed. If no more messages are available then this method will block. It is intended that the client should re-call this method as soon as it has processed the messages received in the previous call. If the client is able to consume messages more quickly than they become available then each call will initially block, there is no need to delay before re-calling this method.  A datafeed will expire if its unread capacity is reached. A datafeed can only be consumed by one client thread at a time. E.g. polling the datafeed by two threads may lead to messages being delivered out of order.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = agent_api.v4_datafeed_id_read_get(id, session_token, key_manager_token, async_req=True)
            >>> result = thread.get()

            Args:
                id (str): Datafeed ID 
                session_token (str): Session authentication token.
                key_manager_token (str): Key Manager authentication token.

            Keyword Args:
                limit (int): Max No. of messages to return. . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                V4EventList
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['id'] = \
                id
            kwargs['session_token'] = \
                session_token
            kwargs['key_manager_token'] = \
                key_manager_token
            return self.call_with_http_info(**kwargs)

        self.v4_datafeed_id_read_get = _Endpoint(
            settings={
                'response_type': (V4EventList,),
                'auth': [],
                'endpoint_path': '/v4/datafeed/{id}/read',
                'operation_id': 'v4_datafeed_id_read_get',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'id',
                    'session_token',
                    'key_manager_token',
                    'limit',
                ],
                'required': [
                    'id',
                    'session_token',
                    'key_manager_token',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'id':
                        (str,),
                    'session_token':
                        (str,),
                    'key_manager_token':
                        (str,),
                    'limit':
                        (int,),
                },
                'attribute_map': {
                    'id': 'id',
                    'session_token': 'sessionToken',
                    'key_manager_token': 'keyManagerToken',
                    'limit': 'limit',
                },
                'location_map': {
                    'id': 'path',
                    'session_token': 'header',
                    'key_manager_token': 'header',
                    'limit': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__v4_datafeed_id_read_get
        )



import logging
import time
from typing import Optional

from symphony.bdk.core.auth.auth_session import AuthSession
from symphony.bdk.core.config.model.bdk_config import BdkConfig
from symphony.bdk.core.retry import retry
from symphony.bdk.core.retry.strategy import read_datafeed_retry
from symphony.bdk.core.service.datafeed.abstract_datafeed_loop import AbstractDatafeedLoop
from symphony.bdk.core.service.datafeed.exception import EventError
from symphony.bdk.core.service.session.session_service import SessionService
from symphony.bdk.gen.agent_api.datafeed_api import DatafeedApi
from symphony.bdk.gen.agent_model.ack_id import AckId
from symphony.bdk.gen.agent_model.v5_datafeed import V5Datafeed
from symphony.bdk.gen.agent_model.v5_datafeed_create_body import V5DatafeedCreateBody

# Based on the DFv2 default visibility timeout, after which an event is re-queued
EVENT_PROCESSING_MAX_DURATION_SECONDS = 30

# DFv2 API authorizes a maximum length for the tag parameter
DATAFEED_TAG_MAX_LENGTH = 100

logger = logging.getLogger(__name__)


class DatafeedLoopV2(AbstractDatafeedLoop):
    """A class for implementing the datafeed v2 loop service.

        This service will be started by calling :func:`~DatafeedLoopV2.start`.

        On the very first run, the BDK bot will try to retrieve the list of datafeed to which it is listening.
        Since each bot should only listening to just one datafeed, the first datafeed in the list will be used by
        the bot to be listened to. If the retrieved list is empty, the BDK bot will create a new datafeed to listen.

        The BDK bot will listen to this datafeed to get all the received real-time events.

        If this datafeed becomes stale or faulty, the BDK bot will create the new one for listening.

        This service will be stopped by calling :func:`~DatafeedLoopV1.stop`

        If the datafeed service is stopped during a read datafeed call, it has to wait until the last read finish to be
        really stopped
        """

    def __init__(self, datafeed_api: DatafeedApi, session_service: SessionService, auth_session: AuthSession,
                 config: BdkConfig):
        super().__init__(datafeed_api, session_service, auth_session, config)
        self._ack_id = ""
        self._datafeed_id = None
        self._tag = config.bot.username[0:DATAFEED_TAG_MAX_LENGTH]

    async def _prepare_datafeed(self):
        datafeed = await self._retrieve_datafeed()
        if not datafeed:
            datafeed = await self._create_datafeed()
        self._datafeed_id = datafeed.id

    async def _run_loop_iteration(self):
        events = await self._read_datafeed()

        is_run_successful = await self._run_all_listener_tasks(events.events)
        if is_run_successful:
            # updates ack id so that on next call DFv2 knows that events have been processed
            # if not updated, events will be requeued after some time, typically 30s
            self._ack_id = events.ack_id

    async def _run_all_listener_tasks(self, events):
        start = time.time()
        done_tasks = await self._run_listener_tasks(events)
        elapsed = time.time() - start

        if elapsed > EVENT_PROCESSING_MAX_DURATION_SECONDS:
            logging.warning("Events processing took longer than %s seconds, "
                            "this might lead to events being re-queued in datafeed and re-dispatched. "
                            "You might want to consider processing the event in a separated asyncio task if needed.",
                            EVENT_PROCESSING_MAX_DURATION_SECONDS)

        return await self._are_tasks_successful(done_tasks)

    async def _are_tasks_successful(self, tasks):
        success = True
        for task in tasks:
            exception = task.exception()
            if exception:
                if isinstance(exception, EventError):
                    logger.warning("Failed to process events inside %s, "
                                   "will not update ack id, events will be re-queued",
                                   task.get_name(),
                                   exc_info=exception)
                    success = False
                else:
                    logging.debug("Exception occurred inside %s", task.get_name(), exc_info=exception)
        return success

    @retry(retry=read_datafeed_retry)
    async def _read_datafeed(self):
        params = {
            "session_token": await self._auth_session.session_token,
            "key_manager_token": await self._auth_session.key_manager_token,
            "datafeed_id": self._datafeed_id,
            "ack_id": AckId(ack_id=self._ack_id)
        }
        return await self._datafeed_api.read_datafeed(**params)

    async def recreate_datafeed(self):
        await self._delete_datafeed()
        datafeed = await self._create_datafeed()

        self._datafeed_id = datafeed.id
        self._ack_id = ""

    @retry
    async def _retrieve_datafeed(self) -> Optional[V5Datafeed]:
        session_token = await self._auth_session.session_token
        key_manager_token = await self._auth_session.key_manager_token
        datafeeds = await self._datafeed_api.list_datafeed(session_token=session_token,
                                                           key_manager_token=key_manager_token,
                                                           tag=self._tag)
        if datafeeds:
            return datafeeds[0]
        return None

    @retry
    async def _create_datafeed(self) -> V5Datafeed:
        session_token = await self._auth_session.session_token
        key_manager_token = await self._auth_session.key_manager_token

        return await self._datafeed_api.create_datafeed(session_token=session_token,
                                                        key_manager_token=key_manager_token,
                                                        body=V5DatafeedCreateBody(tag=self._tag))

    @retry
    async def _delete_datafeed(self) -> None:
        session_token = await self._auth_session.session_token
        key_manager_token = await self._auth_session.key_manager_token
        await self._datafeed_api.delete_datafeed(datafeed_id=self._datafeed_id,
                                                 session_token=session_token,
                                                 key_manager_token=key_manager_token)



from symphony.bdk.core.auth.auth_session import OboAuthSession
from symphony.bdk.core.client.api_client_factory import ApiClientFactory
from symphony.bdk.core.config.model.bdk_config import BdkConfig
from symphony.bdk.core.service.connection.connection_service import OboConnectionService
from symphony.bdk.core.service.message.message_service import OboMessageService
from symphony.bdk.core.service.signal.signal_service import OboSignalService
from symphony.bdk.core.service.stream.stream_service import OboStreamService
from symphony.bdk.core.service.user.user_service import OboUserService
from symphony.bdk.core.service_factory import OboServiceFactory


class OboServices:
    """Entry point for OBO-enabled services, see
        `the list of OBO-enabled endpoints <https://developers.symphony.com/restapi/reference#obo-enabled-endpoints>`_
    """

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close_clients()

    def __init__(self, config: BdkConfig, obo_session: OboAuthSession):
        """

        :param config: the BDK configuration.
        :param obo_session: the OBO session to use.
        """
        self._config = config
        self._obo_session = obo_session

        self._api_client_factory = ApiClientFactory(config)
        self._service_factory = OboServiceFactory(self._api_client_factory, self._obo_session, self._config)
        self._connection_service = self._service_factory.get_connection_service()
        self._message_service = self._service_factory.get_message_service()
        self._stream_service = self._service_factory.get_stream_service()
        self._user_service = self._service_factory.get_user_service()
        self._signal_service = self._service_factory.get_signal_service()

    def connections(self) -> OboConnectionService:
        """

        :return: a fully initialized OboConnectionService instance.
        """
        return self._connection_service

    def messages(self) -> OboMessageService:
        """

        :return: a fully initialized OboMessageService instance.
        """
        return self._message_service

    def streams(self) -> OboStreamService:
        """

        :return: a fully initialized OboStreamService instance.
        """
        return self._stream_service

    def users(self) -> OboUserService:
        """

        :return: a fully initialized OboUserService instance.
        """
        return self._user_service

    def signals(self) -> OboSignalService:
        """

        :return: a fully initialized OboSignalService instance.
        """
        return self._signal_service

    async def close_clients(self):
        """Close all the existing api clients created by the api client factory.
        """
        await self._api_client_factory.close_clients()



"""Symphony BDK core package

There is an outstanding issue with aiohttp and Python 3.8+ on Windows. This affects our client when using a
proxy on Windows with Python 3.8+. As known workaround, the following code is meant to avoid this issue by detecting the
operating system and setting WindowsSelectorEventLoopPolicy as event loop policy.
"""
import asyncio
import os

if os.name == "nt":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
